#!/usr//local/bin/python2.sh
# -*-mode: python; indent-tabs-mode: nil; py-indent-offset: 4; coding: utf-8 -*-
# From: github.com/napalm255/ansible-selenium/library/selenium_test.py

# (c) 2016, Brad Gibson <napalm255@gmail.com>
#
# This file is a 3rd Party module for Ansible
#
# Ansible is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Ansible is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Ansible.  If not, see <http://www.gnu.org/licenses/>.

"""Ansible Selenium Module."""

DOCUMENTATION = '''
---
module: selenium_test
author: "Brad Gibson"
version_added: "2.3"
short_description: Run selenium tests
requires: [ selenium>=3.0.2 ]
description:
    - Run selenium tests against provided URL.
    - Use Clicks, Typing, Waiting and Assertions.
options:
    url:
        required: true
        description:
            - URL to run selenium tests against.
    browser:
        required: false
        default: "phantomjs"
        choices: [ "phantomjs", "firefox", "chrome" ]
        description:
            - Browser to use for testing.
    browser_executable:
        required: false
        default: ""
    browser_log:
        required: false
        default: ""
    browser_timeout:
        required: false
        default: 30
    browser_service_args:
        required: false
        default: []
    webdriver_executable:
        required: false
        default: ""
    webdriver_log:
        required: false
        default: ""
    width:
        required: false
        default: 1024
        description:
            - Browser screen width.
    height:
        required: false
        default: 768
        description:
            - Browser screen height.
    title:
        required: false
        description:
            - Title to validate after initial load.
    screenshot:
        required: false
        default: false
        description:
            - Enable/Disable screenshots.
    screenshot_when:
        required: false
        default: [ "error" ]
        choices: [ "all", "start", "end", "error" ]
        description:
            - Enable/Disable screenshots.
    screenshot_type:
        required: false
        default: "base64"
        choices: [ "base64", "file" ]
        description:
            - Screenshot format.
    screenshot_path:
        required: false
        default: "/tmp"
        description:
            - Screenshot path.
    screenshot_prefix:
        required: false
        default: "selenium_"
        description:
            - Screenshot file prefix.
    implicit_wait:
        required: false
        default: 20
        description:
            - Implicit wait value when loading webpage.
    explicit_wait:
        required: false
        default: 2
        description:
            - Explicit wait value when loading webpage.
    steps:
        required: true
        description:
            - Steps to perform.
    validate_cert:
        required: false
        default: true
        description:
            - Validate SSL certificate.
'''

EXAMPLES = '''
# run basic check against given url
- selenium_test: url=http://www.python.org
'''

# selenium/webdriver/remote/webdriver.py
_W3C_CAPABILITY_NAMES = frozenset([
    'acceptInsecureCerts',
    'browserName',
    'browserVersion',
    'platformName',
    'pageLoadStrategy',
    'proxy',
    'setWindowRect',
    'timeouts',
    'unhandledPromptBehavior',
])
_OSS_W3C_CONVERSION = {
    'acceptSslCerts': 'acceptInsecureCerts',
    'version': 'browserVersion',
    'platform': 'platformName'
}
# caps['proxy']['proxyType'] = caps['proxy']['proxyType'].lower()

# selenium-3.9.0-py2.7.egg/selenium/webdriver/common/proxy.py
ELTS = ['autodetect', 'ftpProxy', 'httpProxy', 'proxyAutoconfigUrl', 'sslProxy', 'noProxy', 'socksProxy', 'socksUsername', 'socksPassword']
def add_to_capabilities(self, proxyTypeString, capabilities):
    """
    DIRECT = ProxyTypeFactory.make(0, 'DIRECT')  # Direct connection, no proxy (default on Windows).
    MANUAL = ProxyTypeFactory.make(1, 'MANUAL')  # Manual proxy settings (e.g., for httpProxy).
    PAC = ProxyTypeFactory.make(2, 'PAC')  # Proxy autoconfiguration from URL.
    RESERVED_1 = ProxyTypeFactory.make(3, 'RESERVED1')  # Never used.
    AUTODETECT = ProxyTypeFactory.make(4, 'AUTODETECT')  # Proxy autodetection (presumably with WPAD).
    SYSTEM = ProxyTypeFactory.make(5, 'SYSTEM')  # Use system settings (default on Linux).
    UNSPECIFIED = ProxyTypeFactory.make(6, 'UNSPECIFIED')  # Not initialized (for internal use).
    """
    proxy_caps = {}
    proxy_caps['proxyType'] = proxyTypeString
    for elt in ELTS:
        proxy_caps[elt] = getattr(self, elt)
    capabilities['proxy'] = proxy_caps

# pylint: disable = wrong-import-position
import sys,os  # noqa
try:
    if sys.version_info < (3, 0):
        from urlparse import urlparse  # noqa
    else:
        from urllib.parse import urlparse  # noqa
    URLPARSE_INSTALLED_ERR = ''
except ImportError as e:
    URLPARSE_INSTALLED_ERR = 'ERROR: selenium_test.py urllib.parse import urlparse not imported.\n' + str(e)

from ansible.module_utils.basic import AnsibleModule, jsonify  # noqa
try:
    from selenium import webdriver
    from selenium.webdriver.common.keys import Keys
    from selenium.webdriver.common.by import By
    from selenium.webdriver.support.ui import WebDriverWait
    from selenium.webdriver.support import expected_conditions as EC
    from selenium.common.exceptions import NoSuchElementException, TimeoutException
    SELENIUM_INSTALLED_ERR = ''
except ImportError as e:
    SELENIUM_INSTALLED_ERR = 'ERROR: selenium_test.py selenium not imported.\n' + str(e)


class AnsibleSelenium(object):
    """Ansible Selenium Class."""

    def __init__(self, module):
        """Init."""
        self.module = module
        self.arg = lambda: None
        for arg in self.module.params:
            setattr(self.arg, arg, self.module.params[arg])

        self.steps_num = len(self.arg.steps) - 1
        self.result = {'changed': False,
                       'failed': False,
                       'results': {
                           'steps': [],
                           'num': self.steps_num
                       }}
        self.browser_oFd = None
        self.browser = self._browser()

    def __enter__(self):
        """Enter by loading website and return self."""
        # validate url
        url_parsed = urlparse(self.arg.url)
        if url_parsed.scheme not in ['http', 'https']:
            self.failed('invalid url scheme: ' + url_parsed.scheme)
        if not url_parsed.netloc:
            self.failed('invalid url: empty netloc')

        # load browser
        self.browser.get(self.arg.url)
        self.result['browser_closed'] = False

        # validate title
        if not self.arg.title:
            #?mike how is this used? Set it to None if missing or blank?
            self.result['results']['title'] = None
        elif self.arg.title in self.browser.title:
            self.result['results']['title'] = True
        else:
            self.result['results']['title'] = False
            self.failed('title "' +self.browser.title +'" does not contain: ' + self.arg.title)

        # process steps
        self.steps()

        return self

    def __exit__(self, type, value, traceback):
        """Exit by closing and quitting the browser."""
        # pylint: disable = redefined-builtin
        if self.browser_oFd:
            self.browser_oFd.close()
            self.browser_oFd = None
        try: self.browser.close()
        except: pass
        try: self.browser.quit()
        except: pass
        self.result['browser_closed'] = True

    def _browser(self):
        """Select browser and return object."""
        name = self.arg.browser
        if 'phantomjs' in name:
            return self._phantomjs()
        elif 'firefox' in name:
            return self._firefox()
        elif 'chrome' in name:
            return self._chrome()
        else:
            raise RuntimeError("Unrecognized browser " +name)

    def _dRaw_proxy(self):
        # FixMe? socks_proxy
        dRaw = dict()
        # SYSTEM
        dRaw["proxyType"] = "MANUAL"
        if ('no_proxy' in os.environ and os.environ['no_proxy']):
            dRaw['noProxy'] = os.environ['no_proxy']
        else:
            dRaw['noProxy'] = 'localhost,127.0.0.1'

        if os.environ['http_proxy']:
            o = urlparse(os.environ['http_proxy'])
            if o.scheme not in ['http', 'https']:
                self.failed('invalid http_proxy url scheme: ' + o.scheme)
            if not o.netloc:
                self.failed('invalid http_proxy url: empty netloc')
            PROXY_HOST, PROXY_PORT = o.netloc.split(':')
            # 'http://' +
            dRaw['httpProxy'] = PROXY_HOST +':' +PROXY_PORT
        if os.environ['https_proxy']:
            o = urlparse(os.environ['https_proxy'])
            if o.scheme not in ['http', 'https']:
                self.failed('invalid https_proxy url scheme: ' + o.scheme)
            if not o.netloc:
                self.failed('invalid https_proxy url: empty netloc')
            PROXY_HOST, PROXY_PORT = o.netloc.split(':')
            # 'https://' +
            dRaw['sslProxy'] = PROXY_HOST +':' +PROXY_PORT
        return dRaw

    def _phantomjs(self):
        """Use PhantomJS browser."""
# Adding the list of possibilities (integer values) for the "network.proxy.type".
# 0 - Direct connection (or) no proxy.
# 1 - Manual proxy configuration
# 2 - Proxy auto-configuration (PAC).
# 4 - Auto-detect proxy settings.
# 5 - Use system proxy settings.
        from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
        # DesiredCapabilities.PHANTOMJS.copy()
        import warnings
        warnings.filterwarnings("ignore", "Selenium support for PhantomJS has been deprecated.*", UserWarning)
        dCapabilities = {
            "browserName": "phantomjs",
            "browserVersion": "",
            "platform": "ANY",
            "javascriptEnabled": True,
        }
        if not self.arg.browser_executable:
            self.arg.browser_executable = 'phantomjs'
        service_args = self.arg.browser_service_args
        service_args += ['--ssl-protocol=any','--web-security=no']
        if ('http_proxy' in os.environ and os.environ['http_proxy']) or \
           ('https_proxy' in os.environ and os.environ['https_proxy']):
            dRaw = self._dRaw_proxy()
            dCapabilities['proxy'] = dRaw
            # bogus - old
            # --proxy-type=<val> 'http' (default), 'none' (disable completely), or 'socks5'
            service_args += ['--proxy='+dRaw['sslProxy'], '--proxy-type='+'http']
        # --ssl-protocol=TLSv1.2
        # --ssl-certificates-path= the location for custom CA certificates (if none set, uses SSL_CERT_DIR)
        if not self.arg.validate_cert:
            service_args.append('--ignore-ssl-errors=true')
        if False:
            # Starts in 'Remote WebDriver mode' (embedded GhostDriver): '[[<IP>:]<PORT>]' (default '127.0.0.1:8910')
            service_args.append('----webdriver=' +'127.0.0.1:8910')
            # WebDriver Logging Level: (supported: 'ERROR', 'WARN', 'INFO', 'DEBUG') (default 'INFO') (NOTE: needs '--webdriver')
            service_args.append('----webdriver-loglevel' +'INFO')

        if self.arg.webdriver_log:
            service_args += ['--webdriver-logfile='+self.arg.webdriver_log]

        # service_args : A List of command line arguments to pass to PhantomJS
        driver = webdriver.PhantomJS(service_args=service_args,
                                     desired_capabilities=dCapabilities)
        driver.set_window_size(self.arg.width, self.arg.height)
        driver.set_page_load_timeout(self.arg.implicit_wait)
        return driver

    def _firefox(self):
        """Use Firefox browser."""
        # pylint: disable = no-self-use
        dCapabilities = {
            "browserName": "firefox",
            "acceptInsecureCerts": True,
        }
        if self.arg.validate_cert:
            dCapabilities["acceptInsecureCerts"] = False

        # from selenium.webdriver.firefox.firefox_profile import FirefoxProfile
        # firefox_profile = FirefoxProfile(profile_directory=None)
        # warnings.warn("Please use FirefoxOptions to set browser profile", DeprecationWarning,

        if ('http_proxy' in os.environ and os.environ['http_proxy']) or \
           ('https_proxy' in os.environ and os.environ['https_proxy']):
            dRaw = self._dRaw_proxy()
            from selenium.webdriver.common.proxy import Proxy
            dRaw["proxyType"] = "MANUAL"
            oProxy = Proxy(raw=dRaw)
            #dRawCopy = dRaw.copy()
            #dRawCopy["proxyType"] = "MANUAL".lower()
            #? dCapabilities['proxy'] = dRawCopy
        else:
            oProxy = None

        from selenium.webdriver.firefox.firefox_binary import FirefoxBinary
        if not self.arg.browser_executable:
            # their code looks for firefox or iceweasel
            self.arg.browser_executable = 'firefox'
        # should use FirefoxBinary.which() to make sure its on the PATH
        self.arg.browser_executable = which(self.arg.browser_executable)
        assert self.arg.browser_executable

        # A file object to redirect the firefox process output to. It can be sys.stdout.
        # morons: not a file
        if not self.arg.browser_log:
            self.browser_oFd = None
        else:
            self.browser_oFd = open(self.arg.browser_log, 'w')

        firefox_binary = FirefoxBinary(firefox_path=self.arg.browser_executable, log_file=self.browser_oFd)
        # firefox_binary.add_command_line_options()

        if self.arg.webdriver_executable == 'legacy':
            dCapabilities["marionette"] = False
            self.arg.webdriver_executable = ''
        else:
            dCapabilities["marionette"] = True
            if not self.arg.webdriver_executable:
                self.arg.webdriver_executable = 'geckodriver'
            if self.arg.webdriver_log:
                log_path = self.arg.webdriver_log
            else:
                log_path = "geckodriver.log"
            self.arg.webdriver_executable = which(self.arg.webdriver_executable)
            assert self.arg.webdriver_executable

        # options overrides capabilities
        # As some of the options, such as `firefox_profile` and
        # `options.profile` are mutually exclusive, precedence is
        # given from how specific the setting is.  `capabilities` is the
        # least specific keyword argument, followed by `options`,
        # followed by `firefox_binary` and `firefox_profile`.

        #geckodriver
        #  -v                     Log level verbosity (-v for debug and -vv for trace level)
        #  -b, --binary <BINARY>  Path to the Firefox binary
        #  --log <LEVEL>          Set Gecko log level [possible values: fatal, error, warn, info, config, debug, trace]
        service_args = self.arg.browser_service_args

        kwargs = dict(firefox_profile=None,
                       firefox_binary=firefox_binary,
                       timeout=self.arg.browser_timeout,
                       capabilities=dCapabilities,
                       proxy=oProxy,
                       executable_path=self.arg.webdriver_executable,
                       options=None,
                       log_path=log_path,
                       firefox_options=None,
                       service_args=service_args)
        driver = webdriver.Firefox(**kwargs)
        return driver

    def _chrome(self):
        """Use Chrome browser."""
        # pylint: disable = no-self-use
        if not self.arg.webdriver_executable:
            self.arg.webdriver_executable = "chromedriver"
        if not self.arg.browser_executable:
            self.arg.browser_executable = '/opt/chromium-browser/chromium-launcher.sh'
        if not self.arg.browser_service_args:
            self.arg.browser_service_args = None
        if not self.arg.webdriver_log:
            self.arg.webdriver_log = None
        # options get added to desired_capabilities - desired_capabilities.update(options.to_capabilities())
        # warnings.warn('use options instead of chrome_options', DeprecationWarning)
        kwargs = dict(executable_path=self.arg.webdriver_executable,
                      port=0,
                      options=None,
                      service_args=self.arg.browser_service_args,
                      desired_capabilities=None,
                      service_log_path=self.arg.webdriver_log,
                      chrome_options=None)
        # selenium/webdriver/chrome/webdriver.py IGNORES the following arguments
        # proxy=None,       keep_alive=False, file_detector=None, to
        # selenium-3.9.0-py2.7.egg/selenium/webdriver/remote/webdriver.py

        driver = webdriver.Chrome(**kwargs)
        return driver

    def failed(self, msg, step=None):
        """Failed."""
        # self.result['failed'] = True
        self.result['msg'] = msg
        self.result['failed'] = True
        if step:
            step['error'] = True
            step['msg'] = msg
            when = self.arg.screenshot_when
            if 'all' in when or 'error' in when:
                step['screenshot'] = self.screenshot('failed')
            self.result['results']['steps'].append(step)
        else:
            self.result['results']['screenshot'] = self.screenshot('failed')
        # cmd=cmd, rc=rc, stdout=out, stderr=err,
        self.module.fail_json(**self.result)

    def screenshot(self, suffix='default'):
        """Screenshot."""
        details = {}
        if 'base64' in self.arg.screenshot_type:
            base64 = self.browser.get_screenshot_as_base64()
            details['base64'] = base64
        elif 'file' in self.arg.screenshot_type:
            path = '%s/%s%s.png' % (self.arg.screenshot_path,
                                    self.arg.screenshot_prefix,
                                    suffix)
            self.browser.get_screenshot_as_file(path)
            details['file'] = path
        return details

    def keys(self, step, step_result):
        """Keys."""
        step_result['keys'] = False
        try:
            keys_method = getattr(self.browser, step['keys']['type'])
            if 'text' in step['keys']:
                value = step['keys']['text']
                assert value and value.lower() != 'undefined'
                keys_method(step['keys']['value']).send_keys(value)
            if 'key' in step['keys']:
                key_type = getattr(Keys, step['keys']['key'])
                assert key_type and key_type.lower() != 'undefined'
                keys_method(step['keys']['value']).send_keys(key_type)
        except KeyError:
            self.failed('configuration failure. check syntax.', step_result)
        except AttributeError:
            self.failed('type error. check syntax.', step_result)
        except NoSuchElementException:
            self.failed('no such element.', step_result)
        step_result['keys'] = True
        return step_result

    def click(self, step, step_result):
        """Click."""
        step_result['click'] = False
        try:
            click_method = getattr(self.browser, step['click']['type'])
            click_method(step['click']['text']).click()
        except KeyError:
            self.failed('KeyError click configuration failure. check syntax.', step_result)
        except AttributeError:
            self.failed('AttributeError. check syntax.', step_result)
        except NoSuchElementException:
            self.failed('no such element.', step_result)
        step_result['click'] = True
        return step_result

    def wait_for(self, step, step_result):
        """Wait for."""
        step_result['wait_for'] = False
        try:
            waitfor_method = getattr(EC, step['wait_for']['method'])
            waitfor_type = getattr(By, step['wait_for']['type'])
            waitfor_text = step['wait_for']['text']
        except KeyError:
            self.failed('KeyError wait_for configuration failure. check syntax.', step_result)
        except AttributeError:
            self.failed('method or type error. check syntax.', step_result)

        try:
            WebDriverWait(self.browser, self.arg.explicit_wait).until(
                waitfor_method((waitfor_type, waitfor_text))
            )
        except TimeoutException:
            self.failed('failure waiting for element.', step_result)
        step_result['wait_for'] = True
        return step_result

    def asserts(self, step, step_result):
        """Assertions.
            /usr/lib64/python2.7/site-packages/selenium/webdriver/remote/webdriver.py
            def find_element_by_id(self, id_):
            def find_elements_by_id(self, id_):
            def find_element_by_xpath(self, xpath):
            def find_elements_by_xpath(self, xpath):
            def find_element_by_link_text(self, link_text):
            def find_elements_by_link_text(self, text):
            def find_element_by_partial_link_text(self, link_text):
            def find_elements_by_partial_link_text(self, link_text):
            def find_element_by_name(self, name):
            def find_elements_by_name(self, name):
            def find_element_by_tag_name(self, name):
            def find_elements_by_tag_name(self, name):
            def find_element_by_class_name(self, name):
            def find_elements_by_class_name(self, name):
            def find_element_by_css_selector(self, css_selector):
            def find_elements_by_css_selector(self, css_selector):
            def find_element(self, by=By.ID, value=None):
            def find_elements(self, by=By.ID, value=None):
        """
        step_result['assert'] = False
        step_result['assert_results'] = []
        for aidx, item in enumerate(step['assert']):
            step_result['assert_results'].append(False)
            try:
                assert_method = getattr(self.browser, item['type'])
                assert assert_method(item['text'])
            except KeyError:
                self.failed('configuration failure. check syntax.',
                            step_result)
            except NoSuchElementException:
                self.failed('no such element: ' +item['type'] +' - ' +item['text'], step_result)
            step_result['assert_results'][aidx] = True
        step_result['assert'] = True
        return step_result

    def steps(self):
        """Loop through steps."""
        for idx, step in enumerate(self.arg.steps):
            step_result = {'id': idx,
                           'screenshot': 'no'}
            if 'name' in step:
                step_result['name'] = step['name']

            if 'keys' in step:
                step_result.update(self.keys(step, step_result))

            if 'click' in step:
                step_result.update(self.click(step, step_result))

            if 'wait_for' in step:
                step_result.update(self.wait_for(step, step_result))

            if 'assert' in step:
                step_result.update(self.asserts(step, step_result))

            when = self.arg.screenshot_when
            if self.arg.screenshot:
                capture = False
                if 'all' in when:
                    capture = True
                elif 'start' in when and idx is 0:
                    capture = True
                elif 'end' in when and idx is self.steps_num:
                    capture = True

                if capture:
                    suffix = idx
                    if 'name' in step:
                        suffix = '%s_%s' % (idx, step['name'])
                    step_result['screenshot'] = self.screenshot(suffix)
            self.result['results']['steps'].append(step_result)


def which(fname):
    """Returns the fully qualified path by searching Path of the given
    name"""
    for pe in os.environ['PATH'].split(os.pathsep):
        checkname = os.path.join(pe, fname)
        if os.access(checkname, os.X_OK) and not os.path.isdir(checkname):
            return checkname
    return None

def test():
    # https://intoli.com/blog/running-selenium-with-headless-firefox/
    from selenium.webdriver.firefox.firefox_binary import FirefoxBinary
    # Set the MOZ_HEADLESS environment variable which casues Firefox to start in headless mode.
    if not 'MOZ_HEADLESS' in os.environ: os.environ['MOZ_HEADLESS'] = "1"

    # Select your Firefox binary.
    firefox_path = '/usr/bin/firefox'
    assert os.path.isfile(firefox_path)
    binary = FirefoxBinary(firefox_path=firefox_path, log_file=sys.stdout)
    # Start selenium with the configured binary.
    driver = webdriver.Firefox(firefox_binary=binary)

    # Visit this webpage.
    driver.get("http://localhost:3128/")
    heading_element = driver.find_element_by_xpath('//*[@href="doc/"]')
    if heading_element:
        textContent = heading_element.get_property('textContent').strip()
        print('textContent = ' +textContent)
        assert textContent == 'The Polipo manual'
    else:
        print("Heading element not found!")

    driver.get("http://localhost:9090/")
    heading_element = driver.find_element_by_xpath('//*[@href="/+history/FrontPage"]')
    if heading_element:
        textContent = heading_element.get_property('textContent').strip()
        print('textContent = ' +textContent)
        assert textContent == 'History'
    else:
        print("Heading element not found!")

    driver.quit()
    # Ideally, we could instruct Firefox to run in headless mode by including the -headless flag when running the
    # binary with something like
    # binary.add_command_line_options('-headless')
    # However, on current versions of Firefox (up to and including Nightly 58.0a1) running on Windows 10 this
    # flag doesn?t seem to work. Luckily, we can achieve the same effect by setting the MOZ_HEADLESS environment
    # variable either from the command line with set MOZ_HEADLESS=1 or from the python script itself as above.
    return 0

def main():
    """Main."""
    # pylint: disable = too-many-branches

    from ansible.module_utils.basic import _load_params,_ANSIBLE_ARGS
    if len(sys.argv) > 1:
        if not os.path.isfile(sys.argv[1]):
            raise RuntimeError('ERROR: ' +'file not found ' +sys.argv[1])

        # this allows us to leave stdin alone for pdb
        try:
            with open(sys.argv[1], 'rb') as fd:
                _ANSIBLE_ARGS = fd.read()
        except Exception as e:
            raise RuntimeError('ERROR: ' +'file not read ' +repr(sys.argv) +str(e))

    try:
        module = AnsibleModule(
            # check_invalid_arguments=False,
            argument_spec=dict(
                url=dict(type='str', required=True),
                browser=dict(type='str', default='phantomjs',
                             choices=['phantomjs', 'firefox', 'chrome']),
                browser_executable=dict(type='path', default=""),
                browser_log=dict(type='path', default=""),
                browser_timeout=dict(type='int', default=30),
                browser_service_args=dict(type='list', default=[]),
                # For firefox - if this is "legacy" marionette will not be used -
                # The legacy driver provided and maintained by the Selenium project
                # doesn't work for Firefox 48 or higher, and will never work for newer versions of Firefox.
                webdriver_executable=dict(type='path', default=""),
                webdriver_log=dict(type='str', default=""),
                width=dict(type='int', default=1024),
                height=dict(type='int', default=768),
                title=dict(type='str', default=""),
                screenshot=dict(type='bool', default=False),
                screenshot_when=dict(type='list', default=['error']),
                screenshot_type=dict(type='str', default='base64',
                                     choices=['file', 'base64']),
                screenshot_path=dict(type='str', default='/tmp'),
                screenshot_prefix=dict(type='str', default='selenium_'),
                steps=dict(type='list', required=True),
                explicit_wait=dict(type='int', default=2),
                implicit_wait=dict(type='int', default=20),
                validate_cert=dict(type='bool', default=True),
            ),
            supports_check_mode=False
        )
    except Exception as e:
        results = {'failed': True, 'msg': 'ERROR: selenium_test.py AnsibleModule failed ' +str(e)}
        print('\n' +jsonify(results))
        sys.exit(1)

    # check urlparse dependency
    if URLPARSE_INSTALLED_ERR:
        dArgs=dict(error=True, msg=URLPARSE_INSTALLED_ERR)
        module.fail_json(**dArgs)
    # check selenium dependency
    if SELENIUM_INSTALLED_ERR:
        dArgs=dict(error=True, msg=SELENIUM_INSTALLED_ERR)
        module.fail_json(**dArgs)
    try:
        # initiate module
        results = {'failed': True, 'msg': 'ERROR: selenium_test.py something went wrong'}
        with AnsibleSelenium(module) as sel:
            results = sel.result
        module.exit_json(**results)
    except Exception as e:
        results = {'failed': True, 'msg': 'ERROR: selenium_test.py AnsibleSelenium failed ' +str(e)}
        print('\n' +jsonify(results))
        sys.exit(2)

if __name__ == '__main__':
    if len(sys.argv) > 1:
        if sys.argv[1] == 'test':
            i = test()
            sys.exit(i)
    main()

# [ -f selenium_test.json ] || yaml2json.bash < selenium_test.json > selenium_test.json
# python2.bash selenium_test.py selenium_test.json
# old -version 3.141 with manifest
# /usr/local/lib64/python2.7/site-packages/selenium/webdriver/firefox/webdriver.xpi
